Ccccccccccccccccccccccccccccccccccccccccccccccccccccccccc -*- fortran -*-
C 
C       SILO include file for Fortran applications.
C 
C  Programmer:   miller86 Thu Nov  6 18:18:51 PST 2025
C 
C  WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING
C     This file is generated automatically from C header files in the
C     SILO distribution.  Do not make changes to this file manually,
C     as those changes will disappear.
C  WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING
C 
C  This file was generated by the following commands...
C        mv silo.inc silo.inc~
C        cat silo.h silo_f.h | mkinc > silo.inc
C 
C cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc


C...Functions.
      EXTERNAL  DBADDCAOPT
      EXTERNAL  DBADDCCMP
      EXTERNAL  DBADDCOPT
      EXTERNAL  DBADDDOPT
      EXTERNAL  DBADDIAOPT
      EXTERNAL  DBADDICMP
      EXTERNAL  DBADDIOPT
      EXTERNAL  DBADDRCMP
      EXTERNAL  DBADDREGION
      EXTERNAL  DBADDREGIONA
      EXTERNAL  DBADDROPT
      EXTERNAL  DBCALCFL
      EXTERNAL  DBCLOSE
      EXTERNAL  DBCREATE
      EXTERNAL  DBERRLVL
      EXTERNAL  DBERRNO
      EXTERNAL  DBFGETCA
      EXTERNAL  DBFREECAOPT
      EXTERNAL  DBFREEMRGTREE
      EXTERNAL  DBFREEOBJECT
      EXTERNAL  DBFREEOPTLIST
      EXTERNAL  DBGET2DSTRLEN
      EXTERNAL  DBGETCA
      EXTERNAL  DBGETCKSUMS
      EXTERNAL  DBGETCOMPRESS
      EXTERNAL  DBGETCURVE
      EXTERNAL  DBGETDEPWARN
      EXTERNAL  DBGETEMPTYOK
      EXTERNAL  DBGETHDFNMS
      EXTERNAL  DBGETOVRWRT
      EXTERNAL  DBGETQV1
      EXTERNAL  DBINQBLEN
      EXTERNAL  DBINQCA
      EXTERNAL  DBINQDTYP
      EXTERNAL  DBINQFILE
      EXTERNAL  DBINQLEN
      EXTERNAL  DBINQMTYP
      EXTERNAL  DBINQVTYP
      EXTERNAL  DBMKDIR
      EXTERNAL  DBMKMRGTREE
      EXTERNAL  DBMKOBJECT
      EXTERNAL  DBMKOPTLIST
      EXTERNAL  DBMKPTR
      EXTERNAL  DBOPEN
      EXTERNAL  DBPMRGV
      EXTERNAL  DBPUTCA
      EXTERNAL  DBPUTCSGM
      EXTERNAL  DBPUTCSGV
      EXTERNAL  DBPUTCSGZL
      EXTERNAL  DBPUTCURVE
      EXTERNAL  DBPUTDEFVARS
      EXTERNAL  DBPUTFL
      EXTERNAL  DBPUTGRPLMAP
      EXTERNAL  DBPUTMAT
      EXTERNAL  DBPUTMMAT
      EXTERNAL  DBPUTMMESH
      EXTERNAL  DBPUTMRGTREE
      EXTERNAL  DBPUTMSP
      EXTERNAL  DBPUTMVAR
      EXTERNAL  DBPUTPM
      EXTERNAL  DBPUTPV1
      EXTERNAL  DBPUTQM
      EXTERNAL  DBPUTQV
      EXTERNAL  DBPUTQV1
      EXTERNAL  DBPUTUM
      EXTERNAL  DBPUTUV1
      EXTERNAL  DBPUTZL
      EXTERNAL  DBPUTZL2
      EXTERNAL  DBRDVAR
      EXTERNAL  DBRDVARSLICE
      EXTERNAL  DBREGFOPTS
      EXTERNAL  DBRMPTR
      EXTERNAL  DBSET2DSTRLEN
      EXTERNAL  DBSETCKSUMS
      EXTERNAL  DBSETCOMPRESS
      EXTERNAL  DBSETCWR
      EXTERNAL  DBSETDEPWARN
      EXTERNAL  DBSETDIR
      EXTERNAL  DBSETEMPTYOK
      EXTERNAL  DBSETHDFNMS
      EXTERNAL  DBSETOVRWRT
      EXTERNAL  DBSHOWERRORS
      EXTERNAL  DBUNREGAFOPTS
      EXTERNAL  DBUNREGFOPTS
      EXTERNAL  DBWRITE
      EXTERNAL  DBWRITESLICE
      EXTERNAL  DBWRTFL

      INTEGER*4  DBADDCAOPT
      INTEGER*4  DBADDCCMP
      INTEGER*4  DBADDCOPT
      INTEGER*4  DBADDDOPT
      INTEGER*4  DBADDIAOPT
      INTEGER*4  DBADDICMP
      INTEGER*4  DBADDIOPT
      INTEGER*4  DBADDRCMP
      INTEGER*4  DBADDREGION
      INTEGER*4  DBADDREGIONA
      INTEGER*4  DBADDROPT
      INTEGER*4  DBCALCFL
      INTEGER*4  DBCLOSE
      INTEGER*4  DBCREATE
      INTEGER*4  DBERRLVL
      INTEGER*4  DBERRNO
      INTEGER*4  DBFGETCA
      INTEGER*4  DBFREECAOPT
      INTEGER*4  DBFREEMRGTREE
      INTEGER*4  DBFREEOBJECT
      INTEGER*4  DBFREEOPTLIST
      INTEGER*4  DBGET2DSTRLEN
      INTEGER*4  DBGETCA
      INTEGER*4  DBGETCKSUMS
      INTEGER*4  DBGETCOMPRESS
      INTEGER*4  DBGETCURVE
      INTEGER*4  DBGETDEPWARN
      INTEGER*4  DBGETEMPTYOK
      INTEGER*4  DBGETHDFNMS
      INTEGER*4  DBGETOVRWRT
      INTEGER*4  DBGETQV1
      INTEGER*4  DBINQBLEN
      INTEGER*4  DBINQCA
      INTEGER*4  DBINQDTYP
      INTEGER*4  DBINQFILE
      INTEGER*4  DBINQLEN
      INTEGER*4  DBINQMTYP
      INTEGER*4  DBINQVTYP
      INTEGER*4  DBMKDIR
      INTEGER*4  DBMKMRGTREE
      INTEGER*4  DBMKOBJECT
      INTEGER*4  DBMKOPTLIST
      INTEGER*4  DBMKPTR
      INTEGER*4  DBOPEN
      INTEGER*4  DBPMRGV
      INTEGER*4  DBPUTCA
      INTEGER*4  DBPUTCSGM
      INTEGER*4  DBPUTCSGV
      INTEGER*4  DBPUTCSGZL
      INTEGER*4  DBPUTCURVE
      INTEGER*4  DBPUTDEFVARS
      INTEGER*4  DBPUTFL
      INTEGER*4  DBPUTGRPLMAP
      INTEGER*4  DBPUTMAT
      INTEGER*4  DBPUTMMAT
      INTEGER*4  DBPUTMMESH
      INTEGER*4  DBPUTMRGTREE
      INTEGER*4  DBPUTMSP
      INTEGER*4  DBPUTMVAR
      INTEGER*4  DBPUTPM
      INTEGER*4  DBPUTPV1
      INTEGER*4  DBPUTQM
      INTEGER*4  DBPUTQV
      INTEGER*4  DBPUTQV1
      INTEGER*4  DBPUTUM
      INTEGER*4  DBPUTUV1
      INTEGER*4  DBPUTZL
      INTEGER*4  DBPUTZL2
      INTEGER*4  DBRDVAR
      INTEGER*4  DBRDVARSLICE
      INTEGER*4  DBREGFOPTS
      INTEGER*4  DBRMPTR
      INTEGER*4  DBSET2DSTRLEN
      INTEGER*4  DBSETCKSUMS
      INTEGER*4  DBSETCOMPRESS
      INTEGER*4  DBSETCWR
      INTEGER*4  DBSETDEPWARN
      INTEGER*4  DBSETDIR
      INTEGER*4  DBSETEMPTYOK
      INTEGER*4  DBSETHDFNMS
      INTEGER*4  DBSETOVRWRT
      INTEGER*4  DBSHOWERRORS
      INTEGER*4  DBUNREGAFOPTS
      INTEGER*4  DBUNREGFOPTS
      INTEGER*4  DBWRITE
      INTEGER*4  DBWRITESLICE
      INTEGER*4  DBWRTFL


C...Parameters.
      INTEGER*4  DBCSG_ANGLE_PNLA
      INTEGER*4  DBCSG_ANGLE_PPA
      INTEGER*4  DBCSG_BOX_XYXY
      INTEGER*4  DBCSG_BOX_XYZXYZ
      INTEGER*4  DBCSG_CIRCLE_PR
      INTEGER*4  DBCSG_COMPLIMENT
      INTEGER*4  DBCSG_CONE_PNLA
      INTEGER*4  DBCSG_CONE_PPA
      INTEGER*4  DBCSG_CYLINDER_PNLR
      INTEGER*4  DBCSG_CYLINDER_PPR
      INTEGER*4  DBCSG_DIFF
      INTEGER*4  DBCSG_ELLIPSE_PRR
      INTEGER*4  DBCSG_ELLIPSOID_PRRR
      INTEGER*4  DBCSG_HEX_6F
      INTEGER*4  DBCSG_INNER
      INTEGER*4  DBCSG_INTERSECT
      INTEGER*4  DBCSG_LINE_G
      INTEGER*4  DBCSG_LINE_PN
      INTEGER*4  DBCSG_LINE_PP
      INTEGER*4  DBCSG_LINE_X
      INTEGER*4  DBCSG_LINE_Y
      INTEGER*4  DBCSG_ON
      INTEGER*4  DBCSG_OUTER
      INTEGER*4  DBCSG_PLANE_G
      INTEGER*4  DBCSG_PLANE_PN
      INTEGER*4  DBCSG_PLANE_PPP
      INTEGER*4  DBCSG_PLANE_X
      INTEGER*4  DBCSG_PLANE_Y
      INTEGER*4  DBCSG_PLANE_Z
      INTEGER*4  DBCSG_POLYGON_KP
      INTEGER*4  DBCSG_POLYHEDRON_KF
      INTEGER*4  DBCSG_PRISM_5F
      INTEGER*4  DBCSG_PYRAMID_5F
      INTEGER*4  DBCSG_QUADRATIC_G
      INTEGER*4  DBCSG_QUADRIC_G
      INTEGER*4  DBCSG_QUAD_4P
      INTEGER*4  DBCSG_SPHERE_PR
      INTEGER*4  DBCSG_SWEEP
      INTEGER*4  DBCSG_TET_4F
      INTEGER*4  DBCSG_TRI_3P
      INTEGER*4  DBCSG_UNION
      INTEGER*4  DBCSG_XFORM
      INTEGER*4  DBOPT_ALIGN
      INTEGER*4  DBOPT_ALLOWMAT0
      INTEGER*4  DBOPT_ALT_NODENUM_VARS
      INTEGER*4  DBOPT_ALT_ZONENUM_VARS
      INTEGER*4  DBOPT_ASCII_LABEL
      INTEGER*4  DBOPT_BASEINDEX
      INTEGER*4  DBOPT_BLOCKORIGIN
      INTEGER*4  DBOPT_BNDNAMES
      INTEGER*4  DBOPT_CONSERVED
      INTEGER*4  DBOPT_COORDSYS
      INTEGER*4  DBOPT_CYCLE
      INTEGER*4  DBOPT_DISJOINT_MODE
      INTEGER*4  DBOPT_DTIME
      INTEGER*4  DBOPT_EXTENSIVE
      INTEGER*4  DBOPT_EXTENTS
      INTEGER*4  DBOPT_EXTENTS_SIZE
      INTEGER*4  DBOPT_FACETYPE
      INTEGER*4  DBOPT_FIRST
      INTEGER*4  DBOPT_GHOST_NODE_LABELS
      INTEGER*4  DBOPT_GHOST_ZONE_LABELS
      INTEGER*4  DBOPT_GROUPINGNAMES
      INTEGER*4  DBOPT_GROUPINGS
      INTEGER*4  DBOPT_GROUPINGS_SIZE
      INTEGER*4  DBOPT_GROUPNUM
      INTEGER*4  DBOPT_GROUPORIGIN
      INTEGER*4  DBOPT_H5_ALIGN_MIN
      INTEGER*4  DBOPT_H5_ALIGN_VAL
      INTEGER*4  DBOPT_H5_CACHE_NBYTES
      INTEGER*4  DBOPT_H5_CACHE_NELMTS
      INTEGER*4  DBOPT_H5_CACHE_POLICY
      INTEGER*4  DBOPT_H5_CORE_ALLOC_INC
      INTEGER*4  DBOPT_H5_CORE_NO_BACK_STORE
      INTEGER*4  DBOPT_H5_DIRECT_BLOCK_SIZE
      INTEGER*4  DBOPT_H5_DIRECT_BUF_SIZE
      INTEGER*4  DBOPT_H5_DIRECT_MEM_ALIGN
      INTEGER*4  DBOPT_H5_FAM_FILE_OPTS
      INTEGER*4  DBOPT_H5_FAM_SIZE
      INTEGER*4  DBOPT_H5_FAPL_HID_T
      INTEGER*4  DBOPT_H5_FCPL_HID_T
      INTEGER*4  DBOPT_H5_FIC_BUF
      INTEGER*4  DBOPT_H5_FIC_FLAGS
      INTEGER*4  DBOPT_H5_FIC_SIZE
      INTEGER*4  DBOPT_H5_FIRST
      INTEGER*4  DBOPT_H5_LAST
      INTEGER*4  DBOPT_H5_LOG_BUF_SIZE
      INTEGER*4  DBOPT_H5_LOG_NAME
      INTEGER*4  DBOPT_H5_META_BLOCK_SIZE
      INTEGER*4  DBOPT_H5_META_EXTENSION
      INTEGER*4  DBOPT_H5_META_FILE_OPTS
      INTEGER*4  DBOPT_H5_MPIO_COMM
      INTEGER*4  DBOPT_H5_MPIO_INFO
      INTEGER*4  DBOPT_H5_MPIP_NO_GPFS_HINTS
      INTEGER*4  DBOPT_H5_RAW_EXTENSION
      INTEGER*4  DBOPT_H5_RAW_FILE_OPTS
      INTEGER*4  DBOPT_H5_SIEVE_BUF_SIZE
      INTEGER*4  DBOPT_H5_SILO_BLOCK_COUNT
      INTEGER*4  DBOPT_H5_SILO_BLOCK_SIZE
      INTEGER*4  DBOPT_H5_SILO_LOG_STATS
      INTEGER*4  DBOPT_H5_SILO_USE_DIRECT
      INTEGER*4  DBOPT_H5_SMALL_RAW_SIZE
      INTEGER*4  DBOPT_H5_USER_DRIVER_ID
      INTEGER*4  DBOPT_H5_USER_DRIVER_INFO
      INTEGER*4  DBOPT_H5_VFD
      INTEGER*4  DBOPT_HAS_EXTERNAL_ZONES
      INTEGER*4  DBOPT_HIDE_FROM_GUI
      INTEGER*4  DBOPT_HI_OFFSET
      INTEGER*4  DBOPT_LABEL
      INTEGER*4  DBOPT_LAST
      INTEGER*4  DBOPT_LLONGNZNUM
      INTEGER*4  DBOPT_LO_OFFSET
      INTEGER*4  DBOPT_MAJORORDER
      INTEGER*4  DBOPT_MATCOLORS
      INTEGER*4  DBOPT_MATCOUNTS
      INTEGER*4  DBOPT_MATLISTS
      INTEGER*4  DBOPT_MATNAME
      INTEGER*4  DBOPT_MATNAMES
      INTEGER*4  DBOPT_MATNOS
      INTEGER*4  DBOPT_MB_BLOCK_NS
      INTEGER*4  DBOPT_MB_BLOCK_TYPE
      INTEGER*4  DBOPT_MB_EMPTY_COUNT
      INTEGER*4  DBOPT_MB_EMPTY_LIST
      INTEGER*4  DBOPT_MB_FILE_NS
      INTEGER*4  DBOPT_MB_REPR_BLOCK_IDX
      INTEGER*4  DBOPT_MISSING_VALUE
      INTEGER*4  DBOPT_MIXLENS
      INTEGER*4  DBOPT_MMESH_NAME
      INTEGER*4  DBOPT_MRGTREE_NAME
      INTEGER*4  DBOPT_MRGV_ONAMES
      INTEGER*4  DBOPT_MRGV_RNAMES
      INTEGER*4  DBOPT_NGROUPS
      INTEGER*4  DBOPT_NMAT
      INTEGER*4  DBOPT_NMATNOS
      INTEGER*4  DBOPT_NMATSPEC
      INTEGER*4  DBOPT_NODENUM
      INTEGER*4  DBOPT_NSPACE
      INTEGER*4  DBOPT_ORIGIN
      INTEGER*4  DBOPT_PHZONELIST
      INTEGER*4  DBOPT_PLANAR
      INTEGER*4  DBOPT_REFERENCE
      INTEGER*4  DBOPT_REGION_PNAMES
      INTEGER*4  DBOPT_REGNAMES
      INTEGER*4  DBOPT_SPECCOLORS
      INTEGER*4  DBOPT_SPECNAMES
      INTEGER*4  DBOPT_TENSOR_RANK
      INTEGER*4  DBOPT_TIME
      INTEGER*4  DBOPT_TOPO_DIM
      INTEGER*4  DBOPT_TV_CONNECTIVITY
      INTEGER*4  DBOPT_UNITS
      INTEGER*4  DBOPT_USESPECMF
      INTEGER*4  DBOPT_XLABEL
      INTEGER*4  DBOPT_XUNITS
      INTEGER*4  DBOPT_XVARNAME
      INTEGER*4  DBOPT_YLABEL
      INTEGER*4  DBOPT_YUNITS
      INTEGER*4  DBOPT_YVARNAME
      INTEGER*4  DBOPT_ZLABEL
      INTEGER*4  DBOPT_ZONECOUNTS
      INTEGER*4  DBOPT_ZONENAMES
      INTEGER*4  DBOPT_ZONENUM
      INTEGER*4  DBOPT_ZUNITS
      INTEGER*4  DBOPT_ZVARNAME
      INTEGER*4  DB_ABORT
      INTEGER*4  DB_ABUTTING
      INTEGER*4  DB_ALL
      INTEGER*4  DB_ALL_AND_DRVR
      INTEGER*4  DB_APPEND
      INTEGER*4  DB_AREA
      INTEGER*4  DB_ARRAY
      INTEGER*4  DB_BLOCKCENT
      INTEGER*4  DB_BNDCENT
      INTEGER*4  DB_CARTESIAN
      INTEGER*4  DB_CHAR
      INTEGER*4  DB_CLOBBER
      INTEGER*4  DB_COLLINEAR
      INTEGER*4  DB_COLMAJOR
      INTEGER*4  DB_COMPAT_OVER_PERF
      INTEGER*4  DB_CONCURRENT
      INTEGER*4  DB_CRAY
      INTEGER*4  DB_CSGMESH
      INTEGER*4  DB_CSGVAR
      INTEGER*4  DB_CSGZONELIST
      INTEGER*4  DB_CURVE
      INTEGER*4  DB_CURVILINEAR
      INTEGER*4  DB_CYLINDRICAL
      INTEGER*4  DB_DEBUG
      INTEGER*4  DB_DEFVARS
      INTEGER*4  DB_DIR
      INTEGER*4  DB_DOUBLE
      INTEGER*4  DB_DOUBLEA
      INTEGER*4  DB_EDGECENT
      INTEGER*4  DB_EDGELIST
      INTEGER*4  DB_EOA
      INTEGER*4  DB_F77NULL
      INTEGER*4  DB_FACECENT
      INTEGER*4  DB_FACELIST
      INTEGER*4  DB_FILE_OPTS_H5_DEFAULT_CORE
      INTEGER*4  DB_FILE_OPTS_H5_DEFAULT_DEFAULT
      INTEGER*4  DB_FILE_OPTS_H5_DEFAULT_DIRECT
      INTEGER*4  DB_FILE_OPTS_H5_DEFAULT_FAMILY
      INTEGER*4  DB_FILE_OPTS_H5_DEFAULT_LOG
      INTEGER*4  DB_FILE_OPTS_H5_DEFAULT_MPIO
      INTEGER*4  DB_FILE_OPTS_H5_DEFAULT_MPIP
      INTEGER*4  DB_FILE_OPTS_H5_DEFAULT_SEC2
      INTEGER*4  DB_FILE_OPTS_H5_DEFAULT_SILO
      INTEGER*4  DB_FILE_OPTS_H5_DEFAULT_SPLIT
      INTEGER*4  DB_FILE_OPTS_H5_DEFAULT_STDIO
      INTEGER*4  DB_FILE_OPTS_LAST
      INTEGER*4  DB_FLOAT
      INTEGER*4  DB_FLOATA
      INTEGER*4  DB_FLOATING
      INTEGER*4  DB_FROMCWR
      INTEGER*4  DB_GHOSTTYPE_INTDUP
      INTEGER*4  DB_GHOSTTYPE_NOGHOST
      INTEGER*4  DB_GROUPELMAP
      INTEGER*4  DB_H5DEFAULT_COMPAT
      INTEGER*4  DB_H5VFD_CORE
      INTEGER*4  DB_H5VFD_DEFAULT
      INTEGER*4  DB_H5VFD_DIRECT
      INTEGER*4  DB_H5VFD_FAMILY
      INTEGER*4  DB_H5VFD_FIC
      INTEGER*4  DB_H5VFD_LOG
      INTEGER*4  DB_H5VFD_MPIO
      INTEGER*4  DB_H5VFD_MPIP
      INTEGER*4  DB_H5VFD_SEC2
      INTEGER*4  DB_H5VFD_SILO
      INTEGER*4  DB_H5VFD_SPLIT
      INTEGER*4  DB_H5VFD_STDIO
      INTEGER*4  DB_H5_FIC_DONT_COPY
      INTEGER*4  DB_H5_FIC_DONT_RELEASE
      INTEGER*4  DB_HDF5
      INTEGER*4  DB_HDF5X
      INTEGER*4  DB_HDF5_CORE
      INTEGER*4  DB_HDF5_CORE_OBSOLETE
      INTEGER*4  DB_HDF5_DIRECT
      INTEGER*4  DB_HDF5_FAMILY
      INTEGER*4  DB_HDF5_LOG
      INTEGER*4  DB_HDF5_MPIO
      INTEGER*4  DB_HDF5_MPIOP
      INTEGER*4  DB_HDF5_MPIOP_OBSOLETE
      INTEGER*4  DB_HDF5_MPIO_OBSOLETE
      INTEGER*4  DB_HDF5_MPIP
      INTEGER*4  DB_HDF5_SEC2
      INTEGER*4  DB_HDF5_SEC2_OBSOLETE
      INTEGER*4  DB_HDF5_SILO
      INTEGER*4  DB_HDF5_SPLIT
      INTEGER*4  DB_HDF5_STDIO
      INTEGER*4  DB_HDF5_STDIO_OBSOLETE
      INTEGER*4  DB_INT
      INTEGER*4  DB_INTA
      INTEGER*4  DB_INTEL
      INTEGER*4  DB_INVALID_OBJECT
      INTEGER*4  DB_LOCAL
      INTEGER*4  DB_LONG
      INTEGER*4  DB_LONG_LONG
      INTEGER*4  DB_MATERIAL
      INTEGER*4  DB_MATSPECIES
      INTEGER*4  DB_MAX_H5_OBJ_VALS
      DOUBLE PRECISION DB_MISSING_VALUE_NOT_SET
      INTEGER*4  DB_MRGTREE
      INTEGER*4  DB_MRGVAR
      INTEGER*4  DB_MULTIBLOCKMESH
      INTEGER*4  DB_MULTIBLOCKVAR
      INTEGER*4  DB_MULTIMAT
      INTEGER*4  DB_MULTIMATSPECIES
      INTEGER*4  DB_MULTIMESH
      INTEGER*4  DB_MULTIMESHADJ
      INTEGER*4  DB_MULTIVAR
      INTEGER*4  DB_NETCDF
      INTEGER*4  DB_NFILES
      INTEGER*4  DB_NFILTERS
      INTEGER*4  DB_NOBJ_TYPES
      INTEGER*4  DB_NOCLOBBER
      INTEGER*4  DB_NODECENT
      INTEGER*4  DB_NONCOLLINEAR
      INTEGER*4  DB_NONE
      INTEGER*4  DB_NOTCENT
      INTEGER*4  DB_NOTYPE
      INTEGER*4  DB_NUMERICAL
      INTEGER*4  DB_OFF
      INTEGER*4  DB_ON
      INTEGER*4  DB_OTHER
      INTEGER*4  DB_PARTIO_HSLABS
      INTEGER*4  DB_PARTIO_POINTS
      INTEGER*4  DB_PDB
      INTEGER*4  DB_PDBP
      INTEGER*4  DB_PERF_OVER_COMPAT
      INTEGER*4  DB_PHZONELIST
      INTEGER*4  DB_POINTMESH
      INTEGER*4  DB_POINTVAR
      INTEGER*4  DB_POSTORDER
      INTEGER*4  DB_PREORDER
      INTEGER*4  DB_QUADCURV
      INTEGER*4  DB_QUADMESH
      INTEGER*4  DB_QUADRECT
      INTEGER*4  DB_QUADVAR
      INTEGER*4  DB_QUAD_CURV
      INTEGER*4  DB_QUAD_RECT
      INTEGER*4  DB_READ
      INTEGER*4  DB_RECTILINEAR
      INTEGER*4  DB_RESUME
      INTEGER*4  DB_ROWMAJOR
      INTEGER*4  DB_RS6000
      INTEGER*4  DB_SGI
      INTEGER*4  DB_SHORT
      INTEGER*4  DB_SPHERICAL
      INTEGER*4  DB_SUN3
      INTEGER*4  DB_SUN4
      INTEGER*4  DB_SUSPEND
      INTEGER*4  DB_SYMLINK
      INTEGER*4  DB_TAURUS
      INTEGER*4  DB_TOP
      INTEGER*4  DB_UCDMESH
      INTEGER*4  DB_UCDVAR
      INTEGER*4  DB_UNKNOWN
      INTEGER*4  DB_USERDEF
      INTEGER*4  DB_VARIABLE
      INTEGER*4  DB_VARTYPE_ARRAY
      INTEGER*4  DB_VARTYPE_LABEL
      INTEGER*4  DB_VARTYPE_MATERIAL
      INTEGER*4  DB_VARTYPE_SCALAR
      INTEGER*4  DB_VARTYPE_SPECIES
      INTEGER*4  DB_VARTYPE_SYMTENSOR
      INTEGER*4  DB_VARTYPE_TENSOR
      INTEGER*4  DB_VARTYPE_VECTOR
      INTEGER*4  DB_VOLUME
      INTEGER*4  DB_ZONECENT
      INTEGER*4  DB_ZONELIST
      INTEGER*4  DB_ZONETYPE_BEAM
      INTEGER*4  DB_ZONETYPE_CONVEX_HULL
      INTEGER*4  DB_ZONETYPE_HEX
      INTEGER*4  DB_ZONETYPE_POLYGON
      INTEGER*4  DB_ZONETYPE_POLYHEDRON
      INTEGER*4  DB_ZONETYPE_PRISM
      INTEGER*4  DB_ZONETYPE_PYRAMID
      INTEGER*4  DB_ZONETYPE_QUAD
      INTEGER*4  DB_ZONETYPE_QUAD_BEAM
      INTEGER*4  DB_ZONETYPE_QUAD_HEX
      INTEGER*4  DB_ZONETYPE_QUAD_PRISM
      INTEGER*4  DB_ZONETYPE_QUAD_PYRAMID
      INTEGER*4  DB_ZONETYPE_QUAD_QUAD
      INTEGER*4  DB_ZONETYPE_QUAD_TET
      INTEGER*4  DB_ZONETYPE_QUAD_TRIANGLE
      INTEGER*4  DB_ZONETYPE_TET
      INTEGER*4  DB_ZONETYPE_TRIANGLE
      INTEGER*4  SILO_VERS_MAJ
      INTEGER*4  SILO_VERS_MIN
      INTEGER*4  SILO_VERS_PAT
      INTEGER*4  SILO_VERS_PRE
      CHARACTER*10  DB_F77NULLSTRING

      PARAMETER (SILO_VERS_MAJ=4)
      PARAMETER (SILO_VERS_MIN=12)
      PARAMETER (SILO_VERS_PAT=0)
      PARAMETER (SILO_VERS_PRE=1)
      PARAMETER (DB_NETCDF=0)
      PARAMETER (DB_PDB=2)
      PARAMETER (DB_TAURUS=3)
      PARAMETER (DB_UNKNOWN=5)
      PARAMETER (DB_DEBUG=6)
      PARAMETER (DB_HDF5X=7)
      PARAMETER (DB_PDBP=1)
      PARAMETER (DB_HDF5_SEC2_OBSOLETE=256)
      PARAMETER (DB_HDF5_STDIO_OBSOLETE=512)
      PARAMETER (DB_HDF5_CORE_OBSOLETE=768)
      PARAMETER (DB_HDF5_MPIO_OBSOLETE=1024)
      PARAMETER (DB_HDF5_MPIOP_OBSOLETE=1280)
      PARAMETER (DB_H5VFD_DEFAULT=0)
      PARAMETER (DB_H5VFD_SEC2=1)
      PARAMETER (DB_H5VFD_STDIO=2)
      PARAMETER (DB_H5VFD_CORE=3)
      PARAMETER (DB_H5VFD_LOG=4)
      PARAMETER (DB_H5VFD_SPLIT=5)
      PARAMETER (DB_H5VFD_DIRECT=6)
      PARAMETER (DB_H5VFD_FAMILY=7)
      PARAMETER (DB_H5VFD_MPIO=8)
      PARAMETER (DB_H5VFD_MPIP=9)
      PARAMETER (DB_H5VFD_SILO=10)
      PARAMETER (DB_H5VFD_FIC=11)
      PARAMETER (DB_FILE_OPTS_H5_DEFAULT_DEFAULT=0)
      PARAMETER (DB_FILE_OPTS_H5_DEFAULT_SEC2=1)
      PARAMETER (DB_FILE_OPTS_H5_DEFAULT_STDIO=2)
      PARAMETER (DB_FILE_OPTS_H5_DEFAULT_CORE=3)
      PARAMETER (DB_FILE_OPTS_H5_DEFAULT_LOG=4)
      PARAMETER (DB_FILE_OPTS_H5_DEFAULT_SPLIT=5)
      PARAMETER (DB_FILE_OPTS_H5_DEFAULT_DIRECT=6)
      PARAMETER (DB_FILE_OPTS_H5_DEFAULT_FAMILY=7)
      PARAMETER (DB_FILE_OPTS_H5_DEFAULT_MPIO=8)
      PARAMETER (DB_FILE_OPTS_H5_DEFAULT_MPIP=9)
      PARAMETER (DB_FILE_OPTS_H5_DEFAULT_SILO=10)
      PARAMETER (DB_FILE_OPTS_LAST=10)
      PARAMETER (DB_HDF5=7)
      PARAMETER (DB_HDF5_SEC2=2055)
      PARAMETER (DB_HDF5_STDIO=4103)
      PARAMETER (DB_HDF5_CORE=6151)
      PARAMETER (DB_HDF5_LOG=8199)
      PARAMETER (DB_HDF5_SPLIT=10247)
      PARAMETER (DB_HDF5_DIRECT=12295)
      PARAMETER (DB_HDF5_FAMILY=14343)
      PARAMETER (DB_HDF5_MPIO=16391)
      PARAMETER (DB_HDF5_MPIOP=18439)
      PARAMETER (DB_HDF5_MPIP=18439)
      PARAMETER (DB_HDF5_SILO=20487)
      PARAMETER (DB_NFILES=256)
      PARAMETER (DB_NFILTERS=32)
      PARAMETER (DB_COLLINEAR=130)
      PARAMETER (DB_NONCOLLINEAR=131)
      PARAMETER (DB_QUAD_RECT=130)
      PARAMETER (DB_QUAD_CURV=131)
      PARAMETER (DB_INVALID_OBJECT=-1)
      PARAMETER (DB_QUADRECT=130)
      PARAMETER (DB_QUADCURV=131)
      PARAMETER (DB_QUADMESH=500)
      PARAMETER (DB_QUADVAR=501)
      PARAMETER (DB_UCDMESH=510)
      PARAMETER (DB_UCDVAR=511)
      PARAMETER (DB_MULTIMESH=520)
      PARAMETER (DB_MULTIVAR=521)
      PARAMETER (DB_MULTIMAT=522)
      PARAMETER (DB_MULTIMATSPECIES=523)
      PARAMETER (DB_MULTIBLOCKMESH=520)
      PARAMETER (DB_MULTIBLOCKVAR=521)
      PARAMETER (DB_MULTIMESHADJ=524)
      PARAMETER (DB_MATERIAL=530)
      PARAMETER (DB_MATSPECIES=531)
      PARAMETER (DB_FACELIST=550)
      PARAMETER (DB_ZONELIST=551)
      PARAMETER (DB_EDGELIST=552)
      PARAMETER (DB_PHZONELIST=553)
      PARAMETER (DB_CSGZONELIST=554)
      PARAMETER (DB_CSGMESH=555)
      PARAMETER (DB_CSGVAR=556)
      PARAMETER (DB_CURVE=560)
      PARAMETER (DB_DEFVARS=565)
      PARAMETER (DB_POINTMESH=570)
      PARAMETER (DB_POINTVAR=571)
      PARAMETER (DB_ARRAY=580)
      PARAMETER (DB_DIR=600)
      PARAMETER (DB_SYMLINK=601)
      PARAMETER (DB_VARIABLE=610)
      PARAMETER (DB_MRGTREE=611)
      PARAMETER (DB_GROUPELMAP=612)
      PARAMETER (DB_MRGVAR=613)
      PARAMETER (DB_USERDEF=700)
      PARAMETER (DB_INT=16)
      PARAMETER (DB_SHORT=17)
      PARAMETER (DB_LONG=18)
      PARAMETER (DB_FLOAT=19)
      PARAMETER (DB_DOUBLE=20)
      PARAMETER (DB_CHAR=21)
      PARAMETER (DB_LONG_LONG=22)
      PARAMETER (DB_INTA=23)
      PARAMETER (DB_FLOATA=24)
      PARAMETER (DB_DOUBLEA=25)
      PARAMETER (DB_NOTYPE=26)
      PARAMETER (DB_CLOBBER=0)
      PARAMETER (DB_NOCLOBBER=1)
      PARAMETER (DB_READ=1)
      PARAMETER (DB_APPEND=2)
      PARAMETER (DB_CONCURRENT=4)
      PARAMETER (DB_COMPAT_OVER_PERF=16)
      PARAMETER (DB_PERF_OVER_COMPAT=32)
      PARAMETER (DB_H5DEFAULT_COMPAT=64)
      PARAMETER (DB_H5_FIC_DONT_COPY=2)
      PARAMETER (DB_H5_FIC_DONT_RELEASE=4)
      PARAMETER (DB_LOCAL=0)
      PARAMETER (DB_SUN3=10)
      PARAMETER (DB_SUN4=11)
      PARAMETER (DB_SGI=12)
      PARAMETER (DB_RS6000=13)
      PARAMETER (DB_CRAY=14)
      PARAMETER (DB_INTEL=15)
      PARAMETER (DBOPT_FIRST=260)
      PARAMETER (DBOPT_ALIGN=260)
      PARAMETER (DBOPT_COORDSYS=262)
      PARAMETER (DBOPT_CYCLE=263)
      PARAMETER (DBOPT_FACETYPE=264)
      PARAMETER (DBOPT_HI_OFFSET=265)
      PARAMETER (DBOPT_LO_OFFSET=266)
      PARAMETER (DBOPT_LABEL=267)
      PARAMETER (DBOPT_XLABEL=268)
      PARAMETER (DBOPT_YLABEL=269)
      PARAMETER (DBOPT_ZLABEL=270)
      PARAMETER (DBOPT_MAJORORDER=271)
      PARAMETER (DBOPT_NSPACE=272)
      PARAMETER (DBOPT_ORIGIN=273)
      PARAMETER (DBOPT_PLANAR=274)
      PARAMETER (DBOPT_TIME=275)
      PARAMETER (DBOPT_UNITS=276)
      PARAMETER (DBOPT_XUNITS=277)
      PARAMETER (DBOPT_YUNITS=278)
      PARAMETER (DBOPT_ZUNITS=279)
      PARAMETER (DBOPT_DTIME=280)
      PARAMETER (DBOPT_USESPECMF=281)
      PARAMETER (DBOPT_XVARNAME=282)
      PARAMETER (DBOPT_YVARNAME=283)
      PARAMETER (DBOPT_ZVARNAME=284)
      PARAMETER (DBOPT_ASCII_LABEL=285)
      PARAMETER (DBOPT_MATNOS=286)
      PARAMETER (DBOPT_NMATNOS=287)
      PARAMETER (DBOPT_MATNAME=288)
      PARAMETER (DBOPT_NMAT=289)
      PARAMETER (DBOPT_NMATSPEC=290)
      PARAMETER (DBOPT_BASEINDEX=291)
      PARAMETER (DBOPT_ZONENUM=292)
      PARAMETER (DBOPT_NODENUM=293)
      PARAMETER (DBOPT_BLOCKORIGIN=294)
      PARAMETER (DBOPT_GROUPNUM=295)
      PARAMETER (DBOPT_GROUPORIGIN=296)
      PARAMETER (DBOPT_NGROUPS=297)
      PARAMETER (DBOPT_MATNAMES=298)
      PARAMETER (DBOPT_EXTENTS_SIZE=299)
      PARAMETER (DBOPT_EXTENTS=300)
      PARAMETER (DBOPT_MATCOUNTS=301)
      PARAMETER (DBOPT_MATLISTS=302)
      PARAMETER (DBOPT_MIXLENS=303)
      PARAMETER (DBOPT_ZONECOUNTS=304)
      PARAMETER (DBOPT_HAS_EXTERNAL_ZONES=305)
      PARAMETER (DBOPT_PHZONELIST=306)
      PARAMETER (DBOPT_MATCOLORS=307)
      PARAMETER (DBOPT_BNDNAMES=308)
      PARAMETER (DBOPT_REGNAMES=309)
      PARAMETER (DBOPT_ZONENAMES=310)
      PARAMETER (DBOPT_HIDE_FROM_GUI=311)
      PARAMETER (DBOPT_TOPO_DIM=312)
      PARAMETER (DBOPT_REFERENCE=313)
      PARAMETER (DBOPT_GROUPINGS_SIZE=314)
      PARAMETER (DBOPT_GROUPINGS=315)
      PARAMETER (DBOPT_GROUPINGNAMES=316)
      PARAMETER (DBOPT_ALLOWMAT0=317)
      PARAMETER (DBOPT_MRGTREE_NAME=318)
      PARAMETER (DBOPT_REGION_PNAMES=319)
      PARAMETER (DBOPT_TENSOR_RANK=320)
      PARAMETER (DBOPT_MMESH_NAME=321)
      PARAMETER (DBOPT_TV_CONNECTIVITY=322)
      PARAMETER (DBOPT_DISJOINT_MODE=323)
      PARAMETER (DBOPT_MRGV_ONAMES=324)
      PARAMETER (DBOPT_MRGV_RNAMES=325)
      PARAMETER (DBOPT_SPECNAMES=326)
      PARAMETER (DBOPT_SPECCOLORS=327)
      PARAMETER (DBOPT_LLONGNZNUM=328)
      PARAMETER (DBOPT_CONSERVED=329)
      PARAMETER (DBOPT_EXTENSIVE=330)
      PARAMETER (DBOPT_MB_FILE_NS=331)
      PARAMETER (DBOPT_MB_BLOCK_NS=332)
      PARAMETER (DBOPT_MB_BLOCK_TYPE=333)
      PARAMETER (DBOPT_MB_EMPTY_LIST=334)
      PARAMETER (DBOPT_MB_EMPTY_COUNT=335)
      PARAMETER (DBOPT_MB_REPR_BLOCK_IDX=336)
      PARAMETER (DBOPT_MISSING_VALUE=337)
      PARAMETER (DBOPT_ALT_ZONENUM_VARS=338)
      PARAMETER (DBOPT_ALT_NODENUM_VARS=339)
      PARAMETER (DBOPT_GHOST_NODE_LABELS=340)
      PARAMETER (DBOPT_GHOST_ZONE_LABELS=341)
      PARAMETER (DBOPT_LAST=499)
      PARAMETER (DBOPT_H5_FIRST=500)
      PARAMETER (DBOPT_H5_VFD=500)
      PARAMETER (DBOPT_H5_RAW_FILE_OPTS=501)
      PARAMETER (DBOPT_H5_RAW_EXTENSION=502)
      PARAMETER (DBOPT_H5_META_FILE_OPTS=503)
      PARAMETER (DBOPT_H5_META_EXTENSION=504)
      PARAMETER (DBOPT_H5_CORE_ALLOC_INC=505)
      PARAMETER (DBOPT_H5_CORE_NO_BACK_STORE=506)
      PARAMETER (DBOPT_H5_META_BLOCK_SIZE=507)
      PARAMETER (DBOPT_H5_SMALL_RAW_SIZE=508)
      PARAMETER (DBOPT_H5_ALIGN_MIN=509)
      PARAMETER (DBOPT_H5_ALIGN_VAL=510)
      PARAMETER (DBOPT_H5_DIRECT_MEM_ALIGN=511)
      PARAMETER (DBOPT_H5_DIRECT_BLOCK_SIZE=512)
      PARAMETER (DBOPT_H5_DIRECT_BUF_SIZE=513)
      PARAMETER (DBOPT_H5_LOG_NAME=514)
      PARAMETER (DBOPT_H5_LOG_BUF_SIZE=515)
      PARAMETER (DBOPT_H5_MPIO_COMM=516)
      PARAMETER (DBOPT_H5_MPIO_INFO=517)
      PARAMETER (DBOPT_H5_MPIP_NO_GPFS_HINTS=518)
      PARAMETER (DBOPT_H5_SIEVE_BUF_SIZE=519)
      PARAMETER (DBOPT_H5_CACHE_NELMTS=520)
      PARAMETER (DBOPT_H5_CACHE_NBYTES=521)
      PARAMETER (DBOPT_H5_CACHE_POLICY=522)
      PARAMETER (DBOPT_H5_FAM_SIZE=523)
      PARAMETER (DBOPT_H5_FAM_FILE_OPTS=524)
      PARAMETER (DBOPT_H5_USER_DRIVER_ID=525)
      PARAMETER (DBOPT_H5_USER_DRIVER_INFO=526)
      PARAMETER (DBOPT_H5_SILO_BLOCK_SIZE=527)
      PARAMETER (DBOPT_H5_SILO_BLOCK_COUNT=528)
      PARAMETER (DBOPT_H5_SILO_LOG_STATS=529)
      PARAMETER (DBOPT_H5_SILO_USE_DIRECT=530)
      PARAMETER (DBOPT_H5_FIC_SIZE=531)
      PARAMETER (DBOPT_H5_FIC_BUF=532)
      PARAMETER (DBOPT_H5_FIC_FLAGS=533)
      PARAMETER (DBOPT_H5_FCPL_HID_T=597)
      PARAMETER (DBOPT_H5_FAPL_HID_T=598)
      PARAMETER (DBOPT_H5_LAST=599)
      PARAMETER (DB_TOP=0)
      PARAMETER (DB_NONE=1)
      PARAMETER (DB_ALL=2)
      PARAMETER (DB_ABORT=3)
      PARAMETER (DB_SUSPEND=4)
      PARAMETER (DB_RESUME=5)
      PARAMETER (DB_ALL_AND_DRVR=6)
      PARAMETER (DB_ROWMAJOR=0)
      PARAMETER (DB_COLMAJOR=1)
      PARAMETER (DB_NOTCENT=0)
      PARAMETER (DB_NODECENT=110)
      PARAMETER (DB_ZONECENT=111)
      PARAMETER (DB_FACECENT=112)
      PARAMETER (DB_BNDCENT=113)
      PARAMETER (DB_EDGECENT=114)
      PARAMETER (DB_BLOCKCENT=115)
      PARAMETER (DB_CARTESIAN=120)
      PARAMETER (DB_CYLINDRICAL=121)
      PARAMETER (DB_SPHERICAL=122)
      PARAMETER (DB_NUMERICAL=123)
      PARAMETER (DB_OTHER=124)
      PARAMETER (DB_RECTILINEAR=100)
      PARAMETER (DB_CURVILINEAR=101)
      PARAMETER (DB_AREA=140)
      PARAMETER (DB_VOLUME=141)
      PARAMETER (DB_ON=1000)
      PARAMETER (DB_OFF=-1000)
      PARAMETER (DB_ABUTTING=142)
      PARAMETER (DB_FLOATING=143)
      PARAMETER (DB_VARTYPE_SCALAR=200)
      PARAMETER (DB_VARTYPE_VECTOR=201)
      PARAMETER (DB_VARTYPE_TENSOR=202)
      PARAMETER (DB_VARTYPE_SYMTENSOR=203)
      PARAMETER (DB_VARTYPE_ARRAY=204)
      PARAMETER (DB_VARTYPE_MATERIAL=205)
      PARAMETER (DB_VARTYPE_SPECIES=206)
      PARAMETER (DB_VARTYPE_LABEL=207)
      PARAMETER (DB_GHOSTTYPE_NOGHOST=0)
      PARAMETER (DB_GHOSTTYPE_INTDUP=1)
      PARAMETER (DBCSG_QUADRIC_G=16777216)
      PARAMETER (DBCSG_SPHERE_PR=33619968)
      PARAMETER (DBCSG_ELLIPSOID_PRRR=33685504)
      PARAMETER (DBCSG_PLANE_G=50331648)
      PARAMETER (DBCSG_PLANE_X=50397184)
      PARAMETER (DBCSG_PLANE_Y=50462720)
      PARAMETER (DBCSG_PLANE_Z=50528256)
      PARAMETER (DBCSG_PLANE_PN=50593792)
      PARAMETER (DBCSG_PLANE_PPP=50659328)
      PARAMETER (DBCSG_CYLINDER_PNLR=67108864)
      PARAMETER (DBCSG_CYLINDER_PPR=67174400)
      PARAMETER (DBCSG_BOX_XYZXYZ=83886080)
      PARAMETER (DBCSG_CONE_PNLA=100663296)
      PARAMETER (DBCSG_CONE_PPA=100728832)
      PARAMETER (DBCSG_POLYHEDRON_KF=117440512)
      PARAMETER (DBCSG_HEX_6F=117506048)
      PARAMETER (DBCSG_TET_4F=117571584)
      PARAMETER (DBCSG_PYRAMID_5F=117637120)
      PARAMETER (DBCSG_PRISM_5F=117702656)
      PARAMETER (DBCSG_QUADRATIC_G=134217728)
      PARAMETER (DBCSG_CIRCLE_PR=150994944)
      PARAMETER (DBCSG_ELLIPSE_PRR=151060480)
      PARAMETER (DBCSG_LINE_G=167772160)
      PARAMETER (DBCSG_LINE_X=167837696)
      PARAMETER (DBCSG_LINE_Y=167903232)
      PARAMETER (DBCSG_LINE_PN=167968768)
      PARAMETER (DBCSG_LINE_PP=168034304)
      PARAMETER (DBCSG_BOX_XYXY=184549376)
      PARAMETER (DBCSG_ANGLE_PNLA=201326592)
      PARAMETER (DBCSG_ANGLE_PPA=201392128)
      PARAMETER (DBCSG_POLYGON_KP=218103808)
      PARAMETER (DBCSG_TRI_3P=218169344)
      PARAMETER (DBCSG_QUAD_4P=218234880)
      PARAMETER (DBCSG_INNER=2130706432)
      PARAMETER (DBCSG_OUTER=2130771968)
      PARAMETER (DBCSG_ON=2130837504)
      PARAMETER (DBCSG_UNION=2130903040)
      PARAMETER (DBCSG_INTERSECT=2130968576)
      PARAMETER (DBCSG_DIFF=2131034112)
      PARAMETER (DBCSG_COMPLIMENT=2131099648)
      PARAMETER (DBCSG_XFORM=2131165184)
      PARAMETER (DBCSG_SWEEP=2131230720)
      PARAMETER (DB_PREORDER=1)
      PARAMETER (DB_POSTORDER=2)
      PARAMETER (DB_FROMCWR=4)
      PARAMETER (DB_F77NULL=(-99))
      PARAMETER (DB_PARTIO_POINTS=1)
      PARAMETER (DB_PARTIO_HSLABS=2)
      PARAMETER (DB_MISSING_VALUE_NOT_SET=1.0D+308)
      PARAMETER (DB_EOA=0)
      PARAMETER (DB_NOBJ_TYPES=25)
      PARAMETER (DB_ZONETYPE_BEAM=10)
      PARAMETER (DB_ZONETYPE_POLYGON=20)
      PARAMETER (DB_ZONETYPE_TRIANGLE=23)
      PARAMETER (DB_ZONETYPE_QUAD=24)
      PARAMETER (DB_ZONETYPE_POLYHEDRON=30)
      PARAMETER (DB_ZONETYPE_TET=34)
      PARAMETER (DB_ZONETYPE_PYRAMID=35)
      PARAMETER (DB_ZONETYPE_PRISM=36)
      PARAMETER (DB_ZONETYPE_HEX=38)
      PARAMETER (DB_ZONETYPE_CONVEX_HULL=40)
      PARAMETER (DB_ZONETYPE_QUAD_BEAM=11)
      PARAMETER (DB_ZONETYPE_QUAD_TRIANGLE=25)
      PARAMETER (DB_ZONETYPE_QUAD_QUAD=26)
      PARAMETER (DB_ZONETYPE_QUAD_TET=118)
      PARAMETER (DB_ZONETYPE_QUAD_PYRAMID=120)
      PARAMETER (DB_ZONETYPE_QUAD_PRISM=113)
      PARAMETER (DB_ZONETYPE_QUAD_HEX=116)
      PARAMETER (DB_MAX_H5_OBJ_VALS=64)
      PARAMETER (DB_F77NULLSTRING="NULLSTRING")


C End.
