#ifndef H5Z_ZFP_LIB_H
#define H5Z_ZFP_LIB_H

#include "H5Zzfp_plugin.h"

#ifdef __cplusplus
extern "C" {
#endif

extern int H5Z_zfp_initialize(void);
extern int H5Z_zfp_finalize(void);

#ifdef __cplusplus
}
#endif

#endif
