#undef DIMS
#undef Scalar
#undef Int
#undef UInt
#undef EBITS
#undef PBITS
#undef NBMASK
#undef TCMASK
#undef FABS
#undef FREXP
#undef LDEXP
#undef TEMPLATE_H
#undef _cat2
#undef _cat3
#undef _t1
#undef _t2
#undef index
