#ifndef VISIT_STREAM_H
#define VISIT_STREAM_H

// Include iostream and some using statements.
#include <iostream>

using std::cerr;
using std::cout;
using std::endl;
using std::ios;
using std::ostream;
using std::streambuf;

// Include fstream and some using statements.
#include <fstream>

using std::ifstream;
using std::istream;
using std::ofstream;
using std::streampos;

// Include strstream and some using statements.
#include <strstream>

using std::ostrstream;

#endif
